@@ -22,6 +22,7 @@
 
 import __main__
 
+from functools import partial
 import sys
 import os
 from urllib.parse import quote as url_quote
@@ -64,15 +65,18 @@ class Quit(RuntimeError):
 
 
 class Timeout(RuntimeError):
-    pass
+
+    def __init__(self, timeout_secs, abort=False):
+        self.timeout_secs = timeout_secs
+        self.abort = abort
 
 
-def alarm_handler(*a):
-    raise Timeout()
+def alarm_handler(to, *a):
+    raise Timeout(to)
 
 
 def timeout_start(to):
-    signal.signal(signal.SIGALRM, alarm_handler)
+    signal.signal(signal.SIGALRM, partial(alarm_handler, to))
     signal.alarm(to)
 
 
@@ -213,44 +217,47 @@ def wait_booted(
     exec_command: List[str],
     timeout: int = 120,
     short_command_timeout: int = 10,
-    long_command_timeout: int = 60,
 ) -> None:
     '''
     Wait until a container or VM with an init system has sufficiently
     booted to interact with it.
 
-    timeout: how long we will continue to poll the container for status
+    timeout: how long to wait for the testbed to complete boot
 
     short_command_timeout: how long we will wait for a relatively quick command
     to finish
-
-    long_command_timeout: how long we will wait for a slower command to finish
     '''
     await_boot = load_shell_script('lib/in-testbed/await-boot.sh')
 
     while timeout > 0:
         timeout -= 1
         time.sleep(1)
-        (rc, out, _) = execute_timeout(
-            None, 10, exec_command + ['runlevel'],
-            stdout=subprocess.PIPE,
-        )
-        if rc != 0:
-            adtlog.debug('wait_booted: polling status failed, retrying...')
-            continue
-        out = out.strip()
-        if out.split()[-1].isdigit():
-            adtlog.debug('waiting for network')
-            check_exec(
-                exec_command + ['sh', '-ec', await_boot],
-                timeout=long_command_timeout,
+        adtlog.debug('wait_booted: trying to execute trivial command on container')
+        try:
+            (rc, _, _) = execute_timeout(
+                None, short_command_timeout, exec_command + ['true'],
             )
-            return
+        except Timeout:
+            # With some virt servers, commands can get stuck when executed
+            # while the testbed is booting, or at least this is what we observe.
+            # The wait_booted() retry loop has its own timeout, so we shouldn't
+            # stop on timeouts internal to the loop.
+            pass
+        if rc == 0:
+            adtlog.debug("wait_booted: trivial command executed successfully")
+            break
+    else:
+        bomb('timed out waiting for testbed to start')
 
-        adtlog.debug('wait_booted: runlevel "%s", retrying...' % out)
-
-    bomb('timed out waiting for testbed to start; '
-         'last runlevel "%s"' % out)
+    adtlog.debug('testbed is alive, waiting for boot to complete')
+    try:
+        check_exec(
+            exec_command + ['sh', '-ec', await_boot],
+            timeout=timeout,
+            fail_on_stderr=False,
+        )
+    except Timeout:
+        bomb('timed out waiting for testbed to complete boot')
 
 
 class timeout:
@@ -360,13 +367,13 @@ def downtmp_mktemp(path):
     fail if they get moved around.
     '''
     if path:
-        check_exec(['mkdir', '--mode=1777', '--parents', downtmp_open],
+        check_exec(['/bin/mkdir', '--mode=1777', '--parents', downtmp_open],
                    downp=True)
         return path
     else:
-        d = check_exec(['mktemp', '--directory', '--tmpdir', 'autopkgtest.XXXXXX'],
+        d = check_exec(['/bin/mktemp', '--directory', '--tmpdir', 'autopkgtest.XXXXXX'],
                        downp=True, outp=True)
-        check_exec(['chmod', '1777', d], downp=True)
+        check_exec(['/bin/chmod', '1777', d], downp=True)
         return d
 
 
@@ -428,10 +435,10 @@ def cmd_reboot(c, ce):
     # save current downtmp; try a few locations, as /var/cache might be r/o
     # (argh Ubuntu touch)
     directories = '/var/cache /home'
-    check_exec(['sh', '-ec', 'for d in %s; do if [ -w $d ]; then '
-                '  tar --warning=none --create --absolute-names '
+    check_exec(['/bin/sh', '-ec', 'for d in %s; do if [ -w $d ]; then '
+                '  /bin/tar --warning=none --create --absolute-names '
                 '''    -f $d/autopkgtest-tmpdir.tar '%s'; '''
-                '  rm -f /run/autopkgtest-reboot-prepare-mark; '
+                '  /bin/rm -f /run/autopkgtest-reboot-prepare-mark; '
                 '  exit 0; fi; done; exit 1' % (directories, downtmp)],
                downp=True, timeout=copy_timeout)
     adtlog.debug('cmd_reboot: saved current downtmp, rebooting')
@@ -450,11 +457,11 @@ def cmd_reboot(c, ce):
     caller.hook_wait_reboot(**wait_reboot_args)
 
     # restore downtmp
-    check_exec(['sh', '-ec', 'for d in %s; do '
+    check_exec(['/bin/sh', '-ec', 'for d in %s; do '
                 'if [ -e $d/autopkgtest-tmpdir.tar ]; then '
-                ' tar --warning=none --extract --absolute-names '
+                ' /bin/tar --warning=none --extract --absolute-names '
                 '     -f $d/autopkgtest-tmpdir.tar;'
-                ' rm $d/autopkgtest-tmpdir.tar; exit 0; '
+                ' /bin/rm $d/autopkgtest-tmpdir.tar; exit 0; '
                 'fi; done; exit 1' % directories],
                downp=True, timeout=copy_timeout)
     adtlog.debug('cmd_reboot: restored downtmp after reboot')
@@ -500,7 +507,7 @@ def copyup_shareddir(tb, host, is_dir, downtmp_host):
             tb_tmp = os.path.join(downtmp, os.path.basename(host))
             adtlog.debug('copyup_shareddir: tb path %s is not already in '
                          'downtmp, copying to %s' % (tb, tb_tmp))
-            check_exec(['cp', '-r', '--preserve=timestamps,links', tb, tb_tmp],
+            check_exec(['/bin/cp', '-r', '--preserve=timestamps,links', tb, tb_tmp],
                        downp=True)
             # translate into host path
             tb = os.path.join(downtmp_host, os.path.basename(host))
@@ -517,7 +524,7 @@ def copyup_shareddir(tb, host, is_dir, downtmp_host):
 
         if tb_tmp:
             adtlog.debug('copyup_shareddir: rm intermediate copy: %s' % tb)
-            check_exec(['rm', '-rf', tb_tmp], downp=True)
+            check_exec(['/bin/rm', '-rf', tb_tmp], downp=True)
     finally:
         timeout_stop()
 
@@ -563,8 +570,8 @@ def copydown_shareddir(host, tb, is_dir, downtmp_host):
         if host == tb:
             host_tmp = None
         else:
-            check_exec(['rm', '-rf', tb], downp=True)
-            check_exec(['cp', '-r', '--preserve=timestamps,links', host, tb],
+            check_exec(['/bin/rm', '-rf', tb], downp=True)
+            check_exec(['/bin/cp', '-r', '--preserve=timestamps,links', host, tb],
                        downp=True)
         if host_tmp:
             (is_dir and shutil.rmtree or os.unlink)(host_tmp)
