@@ -23,7 +23,6 @@
 import os
 import re
 import argparse
-from glob import glob
 
 import adtlog
 import testdesc
@@ -34,16 +33,6 @@ __all__ = ['parse_args']
 actions = None
 
 
-def is_click_src(path):
-    '''Check if path is a click source tree'''
-
-    if os.path.isdir(os.path.join(path, 'click')):
-        return True
-    if glob(os.path.join(path, 'manifest.json*')):
-        return True
-    return False
-
-
 def read_changes(parser, changes):
     '''Expand a Debian .changes file into contained paths'''
 
@@ -64,14 +53,12 @@ def process_package_arguments(parser, args):
     '''Check positional arguments and produce adt_run_args compatible actions list'''
 
     # TODO: This should be simplified now that the old adt_run_args CLI got dropped
-    # Sort action list by deb << dsc and click-source << click, for a "do what
+    # Sort action list by deb << dsc, for a "do what
     # I mean" compatible adt_run_args action list
 
     global actions
     debsrc_action = None
     has_debs = False
-    has_click = False
-    has_clicksrc = False
 
     # expand .changes files
     packages = []
@@ -83,7 +70,7 @@ def process_package_arguments(parser, args):
 
     def set_debsrc(p, kind, built_bin=None):
         nonlocal debsrc_action
-        if has_clicksrc or debsrc_action:
+        if debsrc_action:
             parser.error('You must specify only one source package to test')
         debsrc_action = (kind, p, built_bin)
 
@@ -93,14 +80,6 @@ def process_package_arguments(parser, args):
             has_debs = True
         elif p.endswith('.dsc') and os.path.exists(p):
             set_debsrc(p, 'source')
-        elif p.endswith('.click') and os.path.exists(p):
-            if has_click:
-                parser.error('You must specify at most one tested click package')
-            actions.append(('click', p, None))
-            has_click = True
-        elif is_click_src(p):
-            set_debsrc(p, 'click-source')
-            has_clicksrc = True
         elif re.match('[0-9a-z][0-9a-z.+-]+$', p):
             set_debsrc(p, 'apt-source', False)
         elif os.path.isfile(os.path.join(p, 'debian', 'control')):
@@ -116,34 +95,25 @@ def process_package_arguments(parser, args):
         else:
             parser.error('%s is not a valid test package' % p)
 
-    # translate --installed-click option into an action
-    if args.installed_click:
-        if has_click:
-            parser.error('You must specify at most one tested click package')
-        actions.append(('click', args.installed_click, None))
-        has_click = True
-
     # if no source is given, check if the current directory is a source tree
-    if not debsrc_action and not has_click and os.path.isfile('debian/control'):
+    if not debsrc_action and os.path.isfile('debian/control'):
         if os.path.exists('debian/files'):
             set_debsrc('.', 'built-tree', False)
         else:
             set_debsrc('.', 'unbuilt-tree')
 
-    if not debsrc_action and not has_clicksrc and not has_click:
-        parser.error('You must specify source or click package to test')
+    if not debsrc_action:
+        parser.error('You must specify source package to test')
 
-    if has_debs or has_click:
+    if has_debs:
         args.built_binaries = False
 
     if debsrc_action:
         # some actions above disable built binaries, for the rest use the CLI option
         if debsrc_action[2] is None:
             debsrc_action = (debsrc_action[0], debsrc_action[1], args.built_binaries)
-        if has_clicksrc:
-            actions.insert(0, debsrc_action)
-        else:
-            actions.append(debsrc_action)
+
+        actions.append(debsrc_action)
 
     adtlog.debug('actions: %s' % actions)
     adtlog.debug('build binaries: %s' % args.built_binaries)
@@ -162,6 +132,13 @@ class ArgumentParser(argparse.ArgumentParser):
         return [arg_line.strip()]
 
 
+class AppendCommaSeparatedArg(argparse.Action):
+    def __call__(self, parser, args, value, option_string=None):
+        result = getattr(args, self.dest, [])
+        result.extend(value.split(','))
+        setattr(args, self.dest, result)
+
+
 def parse_args(arglist=None):
     '''Parse autopkgtest command line arguments.
 
@@ -177,11 +154,10 @@ testsrc can be one of a:
  - Debian *.dsc source package
  - Debian *.changes file containing a .dsc source package (and possibly binaries to test)
  - Debian source package directory
- - click source directory (optional if a *.click binary is given whose manifest points to the source)
  - apt source package name (through apt-get source)
  - Debian source package in git (url#branchname)
 
-You can specify local *.deb packages or a single *.click package to test.'''
+You can specify local *.deb packages to test.'''
 
     epilog = '''The -- argument separates the autopkgtest actions and options
 from the virt-server which provides the testbed. See e. g. man autopkgtest-schroot
@@ -196,20 +172,16 @@ for details.'''
     g_test = parser.add_argument_group('arguments for specifying and modifying the test')
     g_test.add_argument('--override-control', metavar='PATH',
                         help='run tests from control file/manifest PATH instead '
-                        'of the source/click package')
-    # Don't display the deprecated argument name in the --help output.
-    g_test.add_argument('--testname', help=argparse.SUPPRESS)
+                        'of the source package')
     g_test.add_argument('--test-name',
-                        dest='testname',
-                        help='run only given test name. '
-                        'This replaces --testname, which is deprecated.')
+                        dest='only_tests', action='append', default=[],
+                        help='run only given test name (repeatable).')
+    g_test.add_argument('--skip-test',
+                        dest='skip_tests', action='append',
+                        help='skip given test name (repeatable).')
     g_test.add_argument('-B', '--no-built-binaries', dest='built_binaries',
                         action='store_false', default=True,
                         help='do not build/use binaries from .dsc, git source, or unbuilt tree')
-    g_test.add_argument('--installed-click', metavar='CLICKNAME',
-                        help='Run tests from already installed click package '
-                        '(e. g. "com.example.myapp"), from specified click '
-                        'source directory or manifest\'s x-source.')
     g_test.add_argument('packages', nargs='*',
                         help='testsrc source package and testbinary packages as above')
 
@@ -244,22 +216,55 @@ for details.'''
     g_setup.add_argument('-U', '--apt-upgrade', dest='setup_commands',
                          action='append_const',
                          const='''(O=$(bash -o pipefail -ec 'apt-get update | tee /proc/self/fd/2') ||'''
-                         '{ [ "${O%404*Not Found*}" = "$O" ] || exit 100; sleep 15; apt-get update; }'''
+                         '{ [ "${O%404*Not Found*}" = "$O" ] || exit 100; sleep 15; apt-get update; }'
                          ' || { sleep 60; apt-get update; } || false)'
-                         ' && $(which eatmydata || true) apt-get dist-upgrade -y -o '
-                         'Dpkg::Options::="--force-confnew"',
+                         ' && $(command -v eatmydata || true) apt-get dist-upgrade -y -o '
+                         'Dpkg::Options::="--force-confnew"'
+                         ' && $(command -v eatmydata || true) apt-get --purge autoremove -y',
                          help='Run apt update/dist-upgrade before the tests')
     g_setup.add_argument('--setup-commands-boot', metavar='COMMANDS_OR_PATH',
                          action='append', default=[],
                          help='Run these commands after --setup-commands, '
                          'and also every time the testbed is rebooted')
+    g_setup.add_argument('--add-apt-source', action='append',
+                         dest='add_apt_sources',
+                         metavar='"deb http://MIRROR SUITE COMPONENT..."',
+                         default=[],
+                         help='Enable additional apt sources')
+    g_setup.add_argument('--add-apt-release', action='append',
+                         dest='add_apt_releases',
+                         metavar='"RELEASE"',
+                         default=[],
+                         help='Add apt source entries for RELEASE')
+    g_setup.add_argument('--pin-packages', action='append',
+                         dest='pin_packages',
+                         metavar='RELEASE=pkgname,src:srcname,...',
+                         default=[],
+                         help='Enable testing with packages from a different release. '
+                         'Sets up apt pinning to use only those packages from RELEASE; '
+                         'src:srcname expands to all binaries of srcname.')
     g_setup.add_argument('--apt-pocket', action='append',
                          metavar='POCKETNAME[=pkgname,src:srcname,...]',
                          default=[],
-                         help='Enable additional apt source for POCKETNAME. '
+                         help='Enable additional apt source for release-POCKETNAME. '
                          'If packages are given, set up apt pinning to use '
-                         'only those packages from POCKETNAME; src:srcname '
+                         'only those packages from release-POCKETNAME; src:srcname '
                          ' expands to all binaries of srcname')
+    g_setup.add_argument('--apt-default-release',
+                         dest='apt_default_release',
+                         metavar='SUITENAME_OR_CODENAME',
+                         help='Define the default release for apt. For apt pinning to work '
+                         'properly, the default release must be set to the release '
+                         'that should provide the packages that are not pinned. For '
+                         'Debian and Ubuntu, this is normally automatically detected '
+                         'from /etc/os-release.')
+    g_setup.add_argument('--no-apt-fallback',
+                         dest='enable_apt_fallback',
+                         action='store_false',
+                         default=True,
+                         help='Disable the apt fallback which is used with --apt-pocket or '
+                         '--pin-packages in case installation of dependencies fails due to '
+                         'strict pinning')
     g_setup.add_argument('--copy', metavar='HOSTFILE:TESTBEDFILE',
                          action='append', default=[],
                          help='Copy file or dir from host into testbed after '
@@ -267,6 +272,11 @@ for details.'''
     g_setup.add_argument('--env', metavar='VAR=value',
                          action='append', default=[],
                          help='Set arbitrary environment variable for builds and test')
+    g_setup.add_argument('--ignore-restrictions', default=[],
+                         metavar='RESTRICTION[,RESTRICTION...]',
+                         action=AppendCommaSeparatedArg,
+                         help='Run tests even if these restrictions would '
+                         'normally prevent it')
 
     # privileges
     g_priv = parser.add_argument_group('user/privilege handling options')
@@ -289,6 +299,9 @@ for details.'''
 
     # timeouts
     g_time = parser.add_argument_group('timeout options')
+    g_time.add_argument(
+        '--timeout', type=int, default=0, help='Global timeout'
+    )
     for k, v in adt_testbed.timeouts.items():
         g_time.add_argument(
             '--timeout-' + k, type=int, dest='timeout_' + k, metavar='T',
@@ -313,6 +326,18 @@ for details.'''
     g_misc.add_argument('--build-parallel', metavar='N',
                         help='Set "parallel=N" DEB_BUILD_OPTION for building '
                         'packages (default: number of available processors)')
+    g_misc.add_argument(
+        '--needs-internet', dest='needs_internet',
+        choices=['run', 'try', 'skip'],
+        default='run',
+        help='Define how to handle the needs-internet restriction. With "try" '
+        'tests with needs-internet restrictions will be run, but if they fail '
+        'they will be treated as flaky tests. With "skip" these tests will be '
+        'skipped immediately and will not be run. With "run" the restriction '
+        'is basically ignored.')
+    g_misc.add_argument(
+        '-V', '--validate', action='store_true', default=False,
+        help='validate the test control file and exit')
     g_misc.add_argument(
         '-h', '--help', action='help', default=argparse.SUPPRESS,
         help='show this help message and exit')
@@ -321,10 +346,6 @@ for details.'''
     file_parser = ArgumentParser(add_help=False)
     arglist = file_parser.parse_known_args(arglist)[1]
 
-    # deprecation warning
-    if '--testname' in arglist:
-        adtlog.warning('--testname is deprecated; use --test-name')
-
     # split off virt-server args
     try:
         sep = arglist.index('--')
@@ -365,6 +386,7 @@ for details.'''
         args.env.append('LANG=' + args.set_lang)
 
     # set (possibly adjusted) timeout defaults
+    adt_testbed.global_timeout = args.timeout
     for k in adt_testbed.timeouts:
         v = getattr(args, 'timeout_' + k)
         if v is None:
@@ -377,20 +399,13 @@ for details.'''
 
     # if we have --setup-commands and it points to a file, read its contents
     for i, c in enumerate(args.setup_commands):
-        # shortcut for shipped scripts
-        if '/' not in c:
-            shipped = os.path.join('/usr/share/autopkgtest/setup-commands', c)
-            if os.path.exists(shipped):
-                c = shipped
+        c = adt_testbed.locate_setup_command(c)
         if os.path.exists(c):
             with open(c, encoding='UTF-8') as f:
                 args.setup_commands[i] = f.read().strip()
 
     for i, c in enumerate(args.setup_commands_boot):
-        if '/' not in c:
-            shipped = os.path.join('/usr/share/autopkgtest/setup-commands', c)
-            if os.path.exists(shipped):
-                c = shipped
+        c = adt_testbed.locate_setup_command(c)
         if os.path.exists(c):
             with open(c, encoding='UTF-8') as f:
                 args.setup_commands_boot[i] = f.read().strip()
